#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.model_selection import train_test_split
from scipy.signal import resample_poly, resample, hilbert, butter, filtfilt
import biosppy
from matplotlib.mlab import specgram
from pywt import downcoef
import librosa
from tensorflow.keras import layers, Model, Input
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.optimizers import Adam, SGD
from tensorflow.keras.models import load_model

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and murmurs.
    if verbose >= 1:
        print('Extracting features and murmurs from the Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)

    envelopes = list()
    murmurs = list()
    outcomes = list()

    N_e = 200
    stride_e = 50

    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)
        num_locations = get_num_locations(current_patient_data)

        recording_information = current_patient_data.split('\n')[1:num_locations+1]
        audible_locations = get_audible_locations(current_patient_data)
        murmur = get_murmur(current_patient_data)
        outcome = get_outcome(current_patient_data)
        sampling_rate = get_frequency(current_patient_data)
        
        for j in range(len(current_recordings)):
            
            entries = recording_information[j].split(' ')
            
            if (entries[0] in audible_locations and murmur == 'Present') or murmur != 'Present':
                    
                # Extract features.
                envelope = get_envelope(current_recordings[j], sampling_rate)
                if murmur == 'Absent':
                    stride = 16000
                    stride_e = 200
                else:
                    stride = 4000
                    stride_e = 50

                for j in range(int(np.floor((envelope.shape[0]-N_e)/stride_e))+1):
                    patch = envelope[int(j*stride_e):int(j*stride_e+N_e),:]
                    envelopes.append(patch)
                    murmurs.append(murmur)
                    outcomes.append(outcome)

    murmurs = np.array(murmurs)
    outcomes = np.array(outcomes)
    envelopes = np.array(envelopes)


    murmurs[np.argwhere(murmurs=='Present')] = 0
    murmurs[np.argwhere(murmurs=='Unknown')] = 1
    murmurs[np.argwhere(murmurs=='Absent')] = 2
    murmurs = murmurs.astype(int)
    
    outcomes[np.argwhere(outcomes=='Abnormal')] = 1
    outcomes[np.argwhere(outcomes=='Normal')] = 0
    outcomes = outcomes.astype(int)

    inputs_e = Input(shape=(None, 4))    
    
    x_e = layers.Bidirectional(layers.LSTM(64,return_sequences=True))(inputs_e)
    x_e = layers.Bidirectional(layers.LSTM(64))(x_e)

    x = layers.Dense(64,activation="relu")(x_e)

    outputs = layers.Dense(3,activation="softmax")(x)
    outputs2 = layers.Dense(1,activation="sigmoid")(x)
    model =Model(inputs=[inputs_e], outputs=[outputs])
    model2 = Model(inputs=[inputs_e], outputs=[outputs2])

    es = EarlyStopping(monitor='val_loss', mode = 'min', verbose = 0, patience = 50)
    mc = ModelCheckpoint('model1.h5', monitor='val_loss', mode='min', verbose = 0, save_best_only=True)

    # Train the model.
    if verbose >= 1:
        print('Training model...')    

    X_etrain, X_eval, y_train, y_val = train_test_split(envelopes, murmurs, test_size=0.1, stratify = murmurs, random_state=1)

    model.compile(loss= ['sparse_categorical_crossentropy','sparse_categorical_crossentropy'],
            optimizer = SGD(learning_rate = 0.001,momentum=0.9), metrics = ['sparse_categorical_crossentropy'])  
    
    model.fit([X_etrain], y_train, validation_data=([X_eval], y_val), batch_size = 512, epochs=300, verbose = 0, callbacks = [es,mc], class_weight = {0:2,1:2,2:1})

    model = load_model('model1.h5')

    X_etrain, X_eval, y_train, y_val = train_test_split(envelopes, outcomes, test_size=0.1, stratify = outcomes, random_state=1)

    mc = ModelCheckpoint('model2.h5', monitor='val_loss', mode='min', verbose = 0, save_best_only=True)
    model2.compile(loss= ['binary_crossentropy','binary_crossentropy'],
            optimizer = SGD(learning_rate = 0.001,momentum=0.9), metrics = ['binary_crossentropy'])

    model2.fit([X_etrain], y_train, validation_data=([X_eval], y_val), batch_size = 512, epochs=300, verbose = 0, callbacks = [es,mc], class_weight = {0:1,1:1})
    model2 = load_model('model2.h5')
    
    # Save the model.
    save_challenge_model(model_folder, murmur_classes, model, outcome_classes, model2)

    if verbose >= 1:
        print('Done.')

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    filename = os.path.join(model_folder, 'model.sav')

    models = list()

    
    murmur_classifier = load_model(os.path.join(model_folder, 'murmur.h5'))
    outcome_classifier = load_model(os.path.join(model_folder, 'outcome.h5'))

    models.append(murmur_classifier)
    models.append(outcome_classifier)
    
    models.append(joblib.load(filename))
    return models

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):

    murmur_classes = model[-1]['murmur_classes']
    murmur_classifier = model[0]
    outcome_classes = model[-1]['outcome_classes']
    outcome_classifier = model[1]

    N_e = 200
    stride_e = 100

    probabilities_murmur = list()
    probabilities_outcome = list()

    for j in range(len(recordings)):

        probabilities = list()
        probabilities2 = list()

        fs = get_frequency(data)
        envelopes = get_envelope(recordings[j], fs)

        for k in range((int(np.floor((envelopes.shape[0]-N_e)/stride_e)))+1):

            patch1 = envelopes[int(k*stride_e):int(k*stride_e+N_e),:].astype('float32')
            probabilities.append(murmur_classifier.predict(np.expand_dims(patch1,axis=0)).ravel())
            probabilities2.append(outcome_classifier.predict(np.expand_dims(patch1,axis=0)).ravel())
            
        probabilities_murmur.append(np.mean(probabilities,axis=0))
        probabilities_outcome.append(np.mean(probabilities2,axis=0))

    most_probable = np.argmax(np.array(probabilities_murmur)[:,0])                  
    murmur_probabilities = probabilities_murmur[most_probable]
    most_probable = np.argmax(np.array(probabilities_outcome))                  
    outcome_probabilities = probabilities_outcome[most_probable]


    # Choose label with highest probability.
    murmur_murmurs = np.zeros(len(murmur_classes), dtype=np.int_)
    idx = np.argmax(murmur_probabilities,axis=0).ravel()
    murmur_murmurs[idx] = 1
    outcome_murmurs = np.zeros(len(outcome_classes), dtype=np.int_)
    if outcome_probabilities >= 0.5:                                                       
        idx = 0
    else:
        idx = 1
    outcome_murmurs[idx] = 1

    outcome_probabilities = np.array([outcome_probabilities[0],1-outcome_probabilities[0]])

    # Concatenate classes, murmurs, and probabilities.
    classes = murmur_classes + outcome_classes
    murmurs = np.concatenate((murmur_murmurs, outcome_murmurs))
    probabilities = np.concatenate((murmur_probabilities, outcome_probabilities))

    return classes, murmurs, probabilities


################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier):
    d = {'murmur_classes': murmur_classes, 'outcome_classes': outcome_classes}
    filename = os.path.join(model_folder, 'model.sav')
    joblib.dump(d, filename, protocol=0)

    filename = os.path.join(model_folder, 'murmur.h5')
    filename2 = os.path.join(model_folder, 'outcome.h5')
    murmur_classifier.save(filename)
    outcome_classifier.save(filename2)

def get_audible_locations(data):
    
    for i, l in enumerate(data.split('\n')):
        try:
            if l.split(' ')[1] == 'locations:':
                locations = l.split(' ')[2].split('+')
        except:
            pass
    return locations


# Extract features from the data.
def homomorphic_envelope(x, fs=1000, f_LPF=8, order=1):

    b, a = butter(order, 2 * f_LPF / fs, 'low')
    he = np.exp(filtfilt(b, a, np.log(np.abs(hilbert(x)))))
    
    # Remove spurious spikes in first sample:
    he[0] = he[1]
    
    return he

def hilbert_envelope(x):
    
    he = np.abs(hilbert(x))    
    
    return he

def psd_envelope(x, fs):
    
    spectrum,frequencies,t = specgram(x,NFFT=int(fs/40),Fs=fs,window=np.hamming(fs/40),noverlap=int(fs/80),mode='psd');

    low_limit_position = np.argwhere(frequencies==40)[0][0]
    high_limit_position = np.argwhere(frequencies==80)[0][0]
    psd = np.mean(spectrum[low_limit_position:high_limit_position+1,:],axis=0);
    
    
    return psd


def wavelet_envelope(x):    
    
    coeffs = downcoef('d',x, 'rbio3.9', level=3)
    coeffs = np.repeat(coeffs,2**3)
    
    d = (len(coeffs) - len(x))/2
    first =  int(np.floor(d))
    last = int(len(coeffs) - np.ceil(d))

    coeffs = coeffs[first:last]
    wavelet_feature = abs(coeffs)
    
    return wavelet_feature
    
    
def spike_removal(signal,fs):
    
    window_size = round(fs/2)
    trailing_samples = len(signal)%window_size

    
    if trailing_samples >  0:
        sampleframes = np.reshape(signal[:-trailing_samples],(window_size,int(len(signal[:-trailing_samples])/window_size)))
    else:
        sampleframes = np.reshape(signal,(window_size,int(len(signal)/window_size)))

        
    MAAs = np.amax(abs(sampleframes),axis=0);

    while len(np.argwhere(MAAs>np.median(MAAs)*3)) > 0:
        
        window_num = np.argmax(MAAs)
        
        if np.size(window_num)>1:
            window_num = window_num[0]

        spike_position = np.argmax(abs(sampleframes[:,window_num]));
        
        if np.size(spike_position)>1:
            spike_position = spike_position[0]
        
        
        zero_crossings = np.concatenate((abs(np.diff(np.sign(sampleframes[:,window_num])))>1,np.array([False])),axis=0)
        try:
            spike_start = max(np.concatenate((np.array([0]), np.argwhere(zero_crossings[0:spike_position+1]==1)[-1])))
        except:
            spike_start = 0
            
        zero_crossings[0:spike_position+1] = 0
        
        try:
            spike_end = min(np.concatenate((np.argwhere(zero_crossings==1)[0], np.array([window_size-1]))))
        except:
            spike_end = window_size -1
            
        sampleframes[spike_start:(spike_end+1),window_num] = 0.0001;    
        
        MAAs = np.amax(abs(sampleframes),axis=0);

    despiked_signal = np.reshape(sampleframes, (np.size(sampleframes),1));
    despiked_signal = np.concatenate((despiked_signal.ravel(),signal[len(despiked_signal):]),axis=0);
    
    return despiked_signal

def get_envelope(signal,fs):
    
    
    # signal = signal[:-fs]
    
    # Filter Design
    order = 2
    passBand = np.array([25, 400])
    
    # Band-Pass filtering of the PCG:        
    filtered,fs,params = biosppy.signals.tools.filter_signal(signal,'butter','bandpass',order,passBand,fs)
    # spike removal
    filtered = spike_removal(np.transpose(filtered),fs)
    filtered = resample(filtered,int((1000/fs)*len(filtered)))

    fs = 1000
    e1 = homomorphic_envelope(filtered,fs)
    downs_e1 = resample_poly(e1,1,fs/50)
    downs_e1, = biosppy.signals.tools.normalize(downs_e1)
    downs_e1[0] = downs_e1[1]

    e2 = hilbert_envelope(filtered)
    downs_e2 = resample_poly(e2,1,fs/50)
    downs_e2, = biosppy.signals.tools.normalize(downs_e2)
    downs_e2[0] = downs_e2[1]

    
    e3 = psd_envelope(filtered, fs)    
    downs_e3 = resample(e3,len(downs_e1))
    downs_e3, = biosppy.signals.tools.normalize(downs_e3)
    downs_e3[0] = downs_e3[1]

    
    e4 = wavelet_envelope(filtered)
    e4 = e4[:len(e1)]
    downs_e4 = resample_poly(e4,1,fs/50)
    downs_e4, = biosppy.signals.tools.normalize(downs_e4)
    downs_e4[0] = downs_e4[1]

    envelopes = np.zeros((len(downs_e1),4))
    envelopes[:,0] = downs_e1
    envelopes[:,1] = downs_e2
    envelopes[:,2] = downs_e3
    envelopes[:,3] = downs_e4
    
    return envelopes
# train_challenge_model("C:/Users/sofia/Documents/example/training_data","C:/Users/sofia/Documents/example/model",1)